// SPDX-License-Identifier: LGPL-3.0-or-later
// See Notices.txt for copyright information
use crate::util::tree_reduce::tree_reduce;
use fayalite::{prelude::*, util::ready_valid::ReadyValid};
use std::{num::NonZeroUsize, ops::Range};

#[hdl_module]
pub fn unit_free_regs_tracker(
    free_at_once: NonZeroUsize,
    alloc_at_once: NonZeroUsize,
    reg_num_width: usize,
) {
    #[hdl]
    let cd: ClockDomain = m.input();
    #[hdl]
    let free_in: Array<ReadyValid<UInt>> =
        m.input(Array[ReadyValid[UInt[reg_num_width]]][free_at_once.get()]);
    #[hdl]
    let alloc_out: Array<ReadyValid<UInt>> =
        m.input(Array[ReadyValid[UInt[reg_num_width]]][alloc_at_once.get()]);
    assert!(
        reg_num_width <= isize::MAX.ilog2() as usize,
        "too many registers"
    );
    let reg_count = 1usize << reg_num_width;
    #[hdl]
    let allocated_reg = reg_builder().clock_domain(cd).reset(vec![false; reg_count]);
    for free_index in 0..free_at_once.get() {
        connect(free_in[free_index].ready, true);
        #[hdl]
        if let HdlSome(free_num) = ReadyValid::firing_data(free_in[free_index]) {
            connect(allocated_reg[free_num], false);
        }
    }
    struct Summary {
        range: Range<usize>,
        count: Expr<UInt>,
        count_overflowed: Expr<Bool>,
        alloc_nums: Expr<Array<UInt>>,
    }
    let count_ty = UInt::range_inclusive(0..=alloc_at_once.get());
    let Some(Summary {
        range: _,
        count,
        count_overflowed,
        alloc_nums,
    }) = tree_reduce(
        (0..reg_count).map(|index| Summary {
            range: index..index + 1,
            count: allocated_reg[index]
                .cast_to_static::<UInt<1>>()
                .cast_to(count_ty),
            count_overflowed: false.to_expr(),
            alloc_nums: repeat(UInt[0].zero(), alloc_at_once.get()),
        }),
        |l, r| {
            let range = l.range.start..r.range.end;
            #[hdl]
            let reduced_count = wire(count_ty);
            connect_any(reduced_count, l.count + r.count);
            #[hdl]
            let reduced_count_overflowed = wire();
            connect(
                reduced_count_overflowed,
                (l.count + r.count).cmp_ne(reduced_count) | l.count_overflowed | r.count_overflowed,
            );
            #[hdl]
            let reduced_alloc_nums = wire(
                Array[UInt[Expr::ty(l.alloc_nums).element().width() + 1]][alloc_at_once.get()],
            );
            for alloc_index in 0..alloc_at_once.get() {
                #[hdl]
                if l.count_overflowed | l.count.cmp_ge(alloc_at_once) {
                    connect(reduced_alloc_nums[alloc_index], l.alloc_nums[alloc_index]);
                } else {
                    connect(reduced_alloc_nums[alloc_index], r.alloc_nums[alloc_index]);
                }
            }
            Summary {
                range,
                count: reduced_count,
                count_overflowed: reduced_count_overflowed,
                alloc_nums: reduced_alloc_nums,
            }
        },
    )
    else {
        unreachable!("reg_count is known to be non-zero");
    };
    for alloc_index in 0..alloc_at_once.get() {
        #[hdl]
        if let HdlSome(alloc_num) = ReadyValid::firing_data(alloc_out[alloc_index]) {
            connect(allocated_reg[alloc_num], true);
        }
        #[hdl]
        if count_overflowed | count.cmp_ge(alloc_index) {
            connect(
                alloc_out[alloc_index].data,
                HdlSome(alloc_nums[alloc_index]),
            );
        } else {
            connect(
                alloc_out[alloc_index].data,
                HdlOption[UInt[reg_num_width]].HdlNone(),
            );
        }
    }
}

// TODO: add formal proof of unit_free_regs_tracker
